<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Render\Element;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\entity_template\Exception\MultipleAvailableBlueprintException;
use Drupal\entity_template\TemplateBuilderManager;
use Drupal\typed_data\Widget\FormWidgetManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * This form allows the user to create entities from an specific builder. This
 * form has 3 steps:
 *
 * 1. The builder selection. The form will show a table of the available
 * builders for the user to select one. The columns of the table are as follows:
 *   - A radio button for user to pick the builder.
 *   - The label of the builder.
 *   - The return type of the builder (hidden if the 'entity_type_id' parameter
 *     is available).
 *
 * As mentioned above, an 'entity_type_id' parameter can be passed in order to
 * filter the builders by return type.
 * The other optional parameter is 'builder_id', which will skip this first step
 * and go directly to the step two.
 *
 * 2. The parameters. If the builder (either selected by the user of given as a
 * parameter) has contexts that need to be entered, the form will show the
 * fields to input these values. The form relies on typed data form widgets to
 * show the correct input for each parameter, if the manager cannot a find an
 * applicable widget for the context, the broken widget will be shown, is up to
 * the developer to create custom widgets for unsupported typed data.
 *
 * If there are no parameters a button will be shown so user can continue with
 * the execution, this button is expected to be automatically submitted by a js
 * script so users don't have to do an extra click.
 *
 * After entering the parameters, the form will attempt to execute the builder,
 * if the execution is successful, the user will be redirected to the canonical
 * page of the last created entity if available, otherwise the page will be
 * reloaded.
 * Notice: A loop can be created here if the following conditions are met:
 *   - A 'builder_id' was passed to the form.
 *   - The builder doesn't have parameters.
 *   - The builder does have only one blueprint.
 *   - The last created entity doesn't have a canonical page.
 *
 * If the execution fails because of multiple blueprints, the user will be taken
 * to step number three.
 *
 * 3. Blueprint selection. A select input field with the available blueprints
 * for the selected builder is shown on the form. The user needs to pick one and
 * continue with the execution which will be the same as in step two.
 *
 * @package Drupal\entity_template_ui\Form
 */
class CreateEntityFromTemplateForm extends FormBase {

  /**
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $templateBuilderManager;

  /**
   * @var \Drupal\typed_data\Widget\FormWidgetManagerInterface
   */
  protected $formWidgetManager;

  /**
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected $typedDataManager;

  /**
   * The message to display upon successful submission. Return NULL to not show
   * any messages.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   */
  protected function getSuccessfulMessage() {
    return $this->t('Builder was successfully executed.');
  }

  /**
   * CreateEntityFromTemplateForm constructor.
   *
   * @param \Drupal\entity_template\TemplateBuilderManager $template_builder_manager
   * @param \Drupal\typed_data\Widget\FormWidgetManagerInterface $form_widget_manager
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typed_data_manager
   */
  public function __construct(
    TemplateBuilderManager $template_builder_manager,
    FormWidgetManagerInterface $form_widget_manager,
    TypedDataManagerInterface $typed_data_manager
  ) {
    $this->templateBuilderManager = $template_builder_manager;
    $this->formWidgetManager = $form_widget_manager;
    $this->typedDataManager = $typed_data_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.builder'),
      $container->get('plugin.manager.typed_data_form_widget'),
      $container->get('typed_data_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'create_entity_from_template';
  }

  /**
   * Form constructor.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   * @param string $entity_type_id
   * @param string $builder_id
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Exception
   */
  public function buildForm(array $form, FormStateInterface $form_state, $entity_type_id = NULL, $builder_id = NULL) {
    if (!$form_state->get('builders') && !$form_state->get('builder')) {
      if ($builder_id && $this->templateBuilderManager->hasDefinition($builder_id)) {
        $builder = $this->templateBuilderManager->createInstance($builder_id);

        $form_state->set('builder', $builder);
        $form_state->set('step', 'parameters');
      }
      else {
        $builders = [];
        foreach ($this->templateBuilderManager->getDefinitions() as $id => $definition) {
          $builder = $this->templateBuilderManager->createInstance($id);

          // If the return type does not have the colon in it, it must be of
          // type entity or entity[]
          $return_type = $builder->getReturnType();
          $return_entity_type_id = '';
          if (strpos($return_type, ':') !== FALSE) {
            list(, $return_entity_type_id) = explode(':', $return_type, 2);
          }

          if (!$entity_type_id || strpos($return_entity_type_id, $entity_type_id) === 0) {
            $builders[$id] = $builder;
          }
        }

        $form_state->set('builders', $builders);
        $form_state->set('step', 'select_builder');
      }
    }

    if ($form_state->get('step') == 'select_builder') {
      $builders = $form_state->get('builders');
      $headers = [
        $this->t('Select'),
        $this->t('Label'),
      ];
      if (!$entity_type_id) {
        $headers[] = $this->t('Builds');
      }
      $form['builders'] = [
        '#type' => 'table',
        '#header' => $headers,
        '#empty' => $this->t('No results.'),
      ];

      $rows = [];
      foreach ($builders as $key => $builder) {
        $row = [
          'select' => [
            '#type' => 'radio',
            '#title' => $this->t('Select'),
            '#title_display' => 'invisible',
            '#parents' => ['builder'],
            '#return_value' => $key,
          ],
          'label' => [
            '#plain_text' => $builder->getPluginDefinition()['label'],
          ],
        ];
        if (!$entity_type_id) {
          $row['builds'] = ['#plain_text' => $builder->getReturnTypeLabel()];
        }
        $rows[] = $row;
      }

      $form['builders'] += $rows;
      $form['select_builder'] = [
        '#type' => 'submit',
        '#value' => $this->t('Select'),
        '#validate' => ['::validateFormSelectBuilder'],
        '#submit' => ['::submitFormSelectBuilder'],
      ];
    }

    if ($form_state->get('step') == 'parameters') {
      $builder = $form_state->get('builder');
      $parameters = $builder->getPluginDefinition()['context_definitions'];

      if (!empty($parameters)) {
        foreach ($parameters as $id => $context_definition) {
          /** @var \Drupal\Core\Plugin\Context\ContextDefinition $context_definition */
          $data_definition = $context_definition->getDataDefinition();
          $typed_data = $this->typedDataManager->createInstance($context_definition->getDataType(), [
            'data_definition' => $data_definition,
            'name' => $id,
            'parent' => null,
          ]);

          /** @var \Drupal\typed_data\Widget\FormWidgetInterface $widget */
          $widget = $this->formWidgetManager->getFormWidget($data_definition);
          $form['params'][$id] = [
            '#widget' => $widget,
            '#typed_data' => $typed_data,
          ];
          $form['params'][$id] += $widget->form(
            $typed_data,
            SubformState::createForSubform($form['params'][$id], $form, $form_state)
          );
        }

        $form['submit'] = [
          '#type' => 'submit',
          '#value' => $this->t('Submit'),
        ];
      }
      else {
        $form['no_params_message'] = [
          '#plain_text' => $this->t('Click on submit if it is not running already.')
        ];
        $form['submit'] = [
          '#type' => 'submit',
          '#value' => $this->t('Submit'),
          '#attributes' => [
            'class' => ['entity-template-ui-trigger-submit']
          ],
        ];
        $form['#attached']['library'][] = 'entity_template_ui/entity_template_ui.trigger_submit';
      }
    }

    if ($form_state->get('step') == 'blueprints') {
      $builder = $form_state->get('builder');
      /** @var \Drupal\entity_template\BlueprintTemplateBlueprintEntityAdaptor[] $blueprints */
      $blueprints = $builder->getAvailableBlueprints();
      $blueprint_options = [];
      foreach ($blueprints as $key => $blueprint) {
        $blueprint_options[$key] = $blueprint->getEntity()->label();
      }

      $form['blueprint_key'] = [
        '#type' => 'select',
        '#title' => $this->t('Select the blueprint'),
        '#options' => $blueprint_options,
        '#required' => TRUE,
      ];

      $form['submit'] = [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
        '#submit' => [
          '::submitFormSelectBlueprint',
          '::submitForm',
        ],
      ];
    }

    return $form;
  }

  /**
   * Validate callback for selecting a builder.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function validateFormSelectBuilder(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getValue('builder')) {
      $form_state->setError($form['builders'], $this->t('Select a builder.'));
    }
  }

  /**
   * Submit callback for selecting a builder.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitFormSelectBuilder(array &$form, FormStateInterface $form_state) {
    $builder = $form_state->get(['builders', $form_state->getValue('builder')]);

    $form_state->set('builder', $builder);
    $form_state->set('step', 'parameters');
    $form_state->setRebuild();
  }

  /**
   * Submit callback for selecting a blueprint.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function submitFormSelectBlueprint(array &$form, FormStateInterface $form_state) {
    $form_state->set('blueprint_key', $form_state->getValue('blueprint_key'));
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    /** @var \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderBase $builder */
    $builder = $form_state->get('builder');

    try {
      // Store the params in the form state to be used again if a blueprint
      // needs to be selected.
      if (!$form_state->get('params') && !empty($form['params'])) {
        $parameters = [];
        foreach (Element::children($form['params']) as $id) {
          /** @var \Drupal\typed_data\Widget\FormWidgetInterface $widget */
          $widget = $form['params'][$id]['#widget'];
          /** @var \Drupal\Core\TypedData\TypedDataInterface $typed_data */
          $typed_data = $form['params'][$id]['#typed_data'];

          $widget->extractFormValues(
            $typed_data,
            SubformState::createForSubform($form['params'][$id], $form, $form_state)
          );
          $parameters[$id] = $typed_data->getValue();
        }

        $form_state->set('params', $parameters);
      }

      /** @var \Drupal\entity_template\BlueprintResult $result */
      $result = $builder->execute($form_state->get('params') ?: [], NULL, $form_state->get('blueprint_key'));
      $items = $result->getItems();
      if (!empty($items)) {
        foreach ($items as $item) {
          $item->save();
        }

        if (!$form_state->getRedirect() && $item->hasLinkTemplate('canonical')) {
          $form_state->setRedirect('entity.'.$item->getEntityTypeId().'.canonical', [
            $item->getEntityTypeId() => $item->id(),
          ]);
        }
      }

      if ($message = $this->getSuccessfulMessage()) {
        $this->messenger()->addStatus($message);
      }
    }
    catch (MultipleAvailableBlueprintException $e) {
      $form_state->set('step', 'blueprints');
      $form_state->setRebuild();
    }
  }

}
